# $Id: Conf.pm,v 1.18 2003/02/27 23:57:24 after Exp $
#
# Copyright (C) 2002 Nicolas Delon
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package Geronimo::Conf;

use strict;
use vars qw/$AUTOLOAD/;
use Geronimo::Netsoul;

sub	OUTPUT_SILENT		{ 0 }
sub	OUTPUT_NORMAL		{ 1 }
sub	OUTPUT_VERBOSE		{ 2 }

my $STRING_PATTERN = q/[^'"\s]+/;	# simple str pattern
my $SQSTRING_PATTERN = q/'[^']*'/;	# simple quoted str pattern
my $DQSTRING_PATTERN = q/"[^"]*"/;	# double quoted str pattern

sub	new
{
    my	$class = shift;
    my	%opt = @_;
    my	$self = { };
    
    bless($self, $class);
    $self->{conf_file} =
    ($^O =~ /win/i ? "geronimo.conf" : "$ENV{HOME}/.geronimo/geronimo.conf");

    $self->{login} = undef;
    $self->{passwd} = undef;

    $self->{ns_host} = undef;
    $self->{ns_port} = undef;
    $self->{local_addr} = undef;
    $self->{reconnect_delay} = undef;

    $self->{nickname} = undef;
    $self->{location} = undef;
    $self->{data} = undef;

    $self->{skin} = undef;
    $self->{resizable} = undef;
    $self->{mail_popup} = undef;
    $self->{mail_beep} = undef;
    $self->{msg_beep} = undef;
    $self->{idle_delay} = undef;
    $self->{output} = undef;
    $self->{logfile} = undef;
    $self->{mua} = undef;
    $self->{term} = undef;

    $self->{show_active} = undef;
    $self->{show_idle} = undef;
    $self->{show_locked} = undef;
    $self->{show_connection} = undef;
    $self->{show_server} = undef;
    $self->{show_unknown} = undef;
    $self->{show_disconnected} = undef;

    $self->{contact_list} = { };
    $self->{ordered_contact_list} = [ ];

    $self->{kill_list} = { };

    return $self;
}

sub	load_directive_outputlevel
{
    my	$self = shift;
    my	$lineno = shift;
    my	$output = shift || "";	# in case 'output' does not have any arguments

    $self->{output} = undef;
    $self->{output} = &OUTPUT_SILENT if ($output =~ /^silent$/i);
    $self->{output} = &OUTPUT_NORMAL if ($output =~ /^normal|default$/i);
    $self->{output} = &OUTPUT_VERBOSE if ($output =~ /^verbose$/i);
    warn "syntax error line $lineno, it should be: "
    . "output [silent|normal|verbose]\n" if (!defined($self->{output}));
}

sub	load_directive_contact
{
    my	$self = shift;
    my	$lineno = shift;
    my	$login = shift || 
	(warn "syntax error line $lineno, it should be: contact login [nickname]\n", return 0);
    my	$nickname = shift || undef;
    my	$h;
    
    $h = { login => $login, nickname => $nickname };
    $self->{contact_list}->{$login} = $h;
    push(@{ $self->{ordered_contact_list} }, $login);
    return 1;
}

sub	load_directive_kill
{
    my	$self = shift;
    my	$lineno = shift;
    my	($login, $msg) = @_;

    if (!defined $login) {
	warn "syntax error line $lineno, it should be: kill login [\"msg\"]\n";
	return 0;
    }
    if (defined $self->{kill_list}->{$login}) {
	warn "login '$login' is present more than one time in the kill list\n";
	return 0;
    }

    $msg = (defined $msg) ? $msg : "[automatic reply] The person you are "
    . "trying to contact can not read your messages, he has killed you";
    $self->{kill_list}->{$login} = {msg => $msg};
    return 1;
}

sub	load_directive_data
{
    my	$self = shift;
    my	$lineno = shift;
    my	$data = shift || 
    (warn "syntax error line $lineno, it should be: data user_data\n",return 0);

    $self->{data} = $data;
    return 1;
}

sub	set_value
{
    my	$self = shift;
    my	$lineno = shift;
    my	$field = shift;
    my	$syntax = shift;
    my	$value_type = shift;
    my	$value = shift || (warn "syntax error line $lineno, syntax: $syntax\n", return 0);
    
    if ($value_type eq 'bool')
    {
	if ($value =~ /^yes|no|on|off|true|false|enable|disable$/i)
	{
	    $self->{$field} = $value =~ /yes|on|true|enable/i;
	}
	else
	{
	    warn "syntax error line $lineno, syntax: $syntax (boolean type)\n";
	    return 0;
	}
    }
    elsif ($value_type eq 'int')
    {
	if ($value =~ /^\d+$/)
	{
	    $self->{$field} = int $value;
	}
	else
	{
	    warn "syntax error line $lineno, syntax: $syntax (integer type)\n";
	    return 0;
	}
    }
    else # string type
    {
	$self->{$field} = $value;
    }
    return 1;
}

sub	load_configfile
{
    my	$self = shift;
    my	@list;
    
    if (!open(FILE, $self->{conf_file})) {
	warn "Warning, can't open " . $self->{conf_file} . ": $!\n";
	# return 1 because the conf file is optional
	return 1;
    }
    while (<FILE>)
    {
	chomp;
	next if (/^\s*$/ or /^\s*\#/);
	@list = /\s*($STRING_PATTERN|$SQSTRING_PATTERN|$DQSTRING_PATTERN)\s*/g;
	# ".*" because values can be empty strings
	s/^(\'|\")(.*)\1$/$2/ foreach (@list);
	$self->load_directive($., @list);
    }
    close FILE;
    return 1;
}

sub	load_environment
{
    my	$self = shift;
    
    $self->{location} = $ENV{NSLOCATION} if (defined $ENV{NSLOCATION});
    $self->{passwd} = $ENV{NSPASSWD} if (defined $ENV{NSPASSWD});
    return 1;
}

sub	load_cmdline
{
    my $self = shift;
    my @cmdline = @_;
    my @opt_list;
    my $arg;
    
    foreach $arg (@cmdline) 
    {
	# warning: the logfile option value can be '-'
    	if ($arg =~ /^-./)
	{
	    $self->load_directive("<cmdline>", @opt_list) if (@opt_list);
	    @opt_list = ();
	    $arg =~ s/^-//;
	}
	push @opt_list, $arg;
    }
    $self->load_directive("<cmdline>", @opt_list) if (@opt_list);
    return 1;
}

sub	load_directive
{
    my	$self = shift;
    my	$lineno = shift;
    my	@list = @_;
    my	%conf_directive = (
	'login' =>		[ 'login your_login', 'str' ],
	'passwd' =>		[ 'passwd your_password', 'str' ],
	'ns_host' =>		[ 'ns_host ns-server', 'str' ],
	'ns_port' =>		[ 'ns_port 4242', 'str' ],
	'local_addr' =>		[ 'local_addr local_host_bind_address', 'str' ],
	'reconnect_delay' =>	[ 'reconnect_delay delay (in seconds)', 'int' ],
	'nickname' =>		[ 'nickname your_nickname', 'str' ],
	'location' =>		[ 'location your_location', 'str' ],
	'data' =>		\&load_directive_data,
	'skin' =>		[ 'skin your_skin', 'str' ],
	'resizable' =>		[ 'resizable yes/no', 'bool' ],
	'mail_popup' =>		[ 'mail_popup enable/disable', 'bool' ],
	'mail_beep' =>		[ 'mail_beep enable/disable', 'bool' ],
	'msg_beep' =>		[ 'msg_beep enable/disable', 'bool' ],
	'idle_delay' =>		[ 'idle_delay delay (in seconds)', 'int' ],
	'output' =>		\&load_directive_outputlevel,
	'logfile' =>		[ 'logfile your_logfile', 'str' ],
	'mua' =>		[ 'mua your_mail_client', 'str' ],
	'term' =>		[ 'term your_X_terminal_emulator', 'str' ],
	'show_active' =>	[ 'show_active yes/no', 'bool' ],
	'show_idle' =>		[ 'show_idle yes/no', 'bool' ],
	'show_locked' =>	[ 'show_locked yes/no', 'bool' ],
	'show_connection' =>	[ 'show_connection yes/no', 'bool' ],
	'show_server' =>	[ 'show_server yes/no', 'bool' ],
	'show_unknown' =>	[ 'show_unknown yes/no', 'bool' ],
	'show_disconnected' =>	[ 'show_disconnected yes/no', 'bool' ],
	'contact' =>		\&load_directive_contact,
	'kill' =>		\&load_directive_kill,
    );
    my	$directive;
    my	$retval;
    
    # at this point @list contains for example: ('show_active', 'on')
    $directive = shift @list;
    if (!defined $conf_directive{$directive}) {
	($lineno =~ /^\d+/)
	? warn "$self->{conf_file}: line $lineno: invalid option '$directive'"
	: warn "invalid option '$directive' in command line arguments";
	# additional infos
	print("please rename 'mail_popup_beep' to 'mail_beep'\n")
	if ($directive eq 'mail_popup_beep');
	print("please rename 'msg_popup_beep' to 'msg_beep'\n")
	if ($directive eq 'msg_popup_beep');
	print("option 'proxy' is depreciated, remove it, it is not useful\n")
	if ($directive eq 'proxy');
	print("please rename 'server' to 'ns_host'\n")
	if ($directive eq 'server');
	print("option 'log' does not exist anymore and the behaviours of "
	    . "'output' and 'logfile' have changed, read the doc\n")
	if ($directive eq 'log');
	print("please rename 'auto_away_delay' to 'idle_delay'\n")
	if ($directive eq 'auto_away_delay');
	print("please rename 'show_away' to 'show_idle'\n")
	if ($directive eq 'show_away');
    } else {
	if (ref $conf_directive{$directive} eq 'ARRAY')
	{
	    $retval = $self->set_value($lineno, $directive,
		@{$conf_directive{$directive}}, @list);
	}
	else
	{
	    $retval = &{$conf_directive{$directive}}($self, $lineno, @list);
	}
    }
}

sub	load
{
    my	$self = shift;
    
    $self->load_configfile and
    $self->load_environment and
    $self->load_cmdline(@ARGV);

    $self->{login} = $ENV{USER} || '' if (!defined $self->{login});
    $self->{passwd} = '' if (!defined $self->{passwd});
    $self->{ns_host} = "ns-server.epita.fr" if (!defined $self->{ns_host});
    $self->{ns_port} = 4242 if (!defined $self->{ns_port});
    $self->{local_addr} = "0.0.0.0" if (!defined $self->{local_addr});
    $self->{reconnect_delay} = 30 if (!defined $self->{reconnect_delay});
    $self->{nickname} = $self->{login} if (!defined $self->{nickname});
    $self->{location} = ($ENV{HOST} or $ENV{HOSTNAME} or $ENV{COMPUTERNAME}
	or "HomeSweetHome") if (!defined $self->{location});
    $self->{data} = "I'm a newbie, and I didn't even read the config file"
	if (!defined $self->{data});
    $self->{skin} = 'dotNET-gray' if (!defined $self->{skin});
    $self->{resizable} = 1 if (!defined $self->{resizable});
    $self->{mail_popup} = 1 if (!defined $self->{mail_popup});
    $self->{mail_beep} = 1 if (!defined $self->{mail_beep});
    $self->{msg_beep} = 1 if (!defined $self->{msg_beep});
    $self->{idle_delay} = 3 * 60 if (!defined $self->{idle_delay});
    $self->{output} = 2 if (!defined $self->{output});
    $self->{logfile} = '-' if (!defined $self->{logfile});
    $self->{mua} = 'mutt' if (!defined $self->{mua});
    $self->{term} = 'xterm' if (!defined $self->{term});
    $self->{show_active} = 1 if (!defined $self->{show_active});;
    $self->{show_idle} = 1 if (!defined $self->{show_idle});;
    $self->{show_locked} = 1 if (!defined $self->{show_locked});
    $self->{show_connection} = 1 if (!defined $self->{show_connection});
    $self->{show_server} = 1 if (!defined $self->{show_server});
    $self->{show_unknown} = 1 if (!defined $self->{show_unknown});
    $self->{show_disconnected} = 1 if (!defined $self->{show_disconnected});
    return 1;
}

sub	print
{
    my	$self = shift;
    my	$contact;
    
    print "configuration file: $self->{conf_file}\n";
    print "login: $self->{login}\n";
    print "nickname: ", $self->{nickname} || "<empty>", "\n";
    print "password: ", $self->{passwd} || "<empty>", "\n";
    print "location: ", $self->{location} || "<empty>", "\n";
    print "skin ", $self->{skin} || "<empty>", "\n";
    foreach (sort keys %{ $self->{contact_list} })
    {
	$contact = $self->{contact_list}->{$_};
	print "contact $contact->{login}";
	print " $contact->{nickname}" if (defined $contact->{nickname});
	print "\n";
    }
}

sub	AUTOLOAD
{
    my	$self = shift;
    my	$var = $AUTOLOAD;
    
    $var =~ s/^Geronimo::Conf:://;
    die "no such function: $var" if (not $var =~ s/^get_//);
    die "no such property: $var" if (not exists $self->{$var});
    return $self->{$var};
}

sub	DESTROY
{
    
}

sub	get_contact_list
{
    %{ (shift)->{contact_list} };
}

sub	get_ordered_contact_list
{
    @{ (shift)->{ordered_contact_list} };
}

sub	get_show
{
    my	$self = shift;
    my	$state = shift;
    my	%hash = (
	&ST_ACTIVE =>		'active',
	&ST_IDLE =>		'idle',
	&ST_LOCK =>		'locked',
	&ST_CONNECTION =>	'connection',
	&ST_SERVER =>		'server',
	&ST_UNKNOWN =>		'unknown',
	&ST_DISCONNECTED =>	'disconnected',
    );

    return (defined $hash{$state} and $self->{"show_$hash{$state}"});
}

1;
