# $Id: Netsoul.pm,v 1.68 2003/03/08 21:01:31 after Exp $
#
# Copyright (C) 2003 Marc Bevand
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

package Geronimo::Netsoul;

BEGIN
{
    use	Exporter;
    
    @ISA = qw/Exporter/;
    @EXPORT = qw/EV_CONN EV_LOGIN EV_LOGOUT EV_MSG EV_STATE EV_TALK EV_TALKANS
    EV_NEWMAIL EV_PING CONN_DOWN CONN_TRY CONN_UP CONN_BANNER CONN_AUTH_REQ
    CONN_AUTH_REP CONN_LOG_REQ CONN_LOG_REP CONN_READY ST_ACTIVE ST_IDLE
    ST_LOCK ST_CONNECTION ST_SERVER ST_UNKNOWN ST_DISCONNECTED MODE_EXTERNAL
    MODE_INTERNAL REQ_RD_START REQ_RD_STOP REQ_WR_START REQ_WR_STOP/;
}

use strict;
use IO::Socket;
use Digest::MD5 qw/md5_hex/;

$Geronimo::Netsoul::VERSION = "0.8.0";

#
# Exported constants
#

# Events
sub	EV_CONN		{ 0 }	# connection state has changed
sub	EV_LOGIN	{ 1 }	# cmd login received
sub	EV_LOGOUT	{ 2 }	# cmd logout received
sub	EV_MSG		{ 3 }	# cmd msg received
sub	EV_STATE	{ 4 }	# cmd state received
sub	EV_TALK		{ 5 }	# cmd talk received
sub	EV_TALKANS	{ 6 }	# cmd talkans received
sub	EV_NEWMAIL	{ 7 }	# cmd new_mail received
sub	EV_PING		{ 8 }	# ping received
# Connection states
sub	CONN_DOWN	{ 0 }	# not connected
sub	CONN_TRY	{ 1 }	# trying to connect
sub	CONN_UP		{ 2 }	# connection established
sub	CONN_BANNER	{ 3 }	# banner received
sub	CONN_AUTH_REQ	{ 4 }	# auth_ag sent
sub	CONN_AUTH_REP	{ 5 }	# auth_ag reply received
sub	CONN_LOG_REQ	{ 6 }	# log cmd sent
sub	CONN_LOG_REP	{ 7 }	# log cmd reply received
sub	CONN_READY	{ 8 }	# connection ready to use
# Logical states
sub	ST_ACTIVE	{ 0 }	# 'actif' state
sub	ST_IDLE		{ 1 }	# 'idle' or 'away' states
sub	ST_LOCK		{ 2 }	# 'lock' state
sub	ST_CONNECTION	{ 3 }	# 'connection' state
sub	ST_SERVER	{ 4 }	# 'server' state
sub	ST_UNKNOWN	{ 5 }	# non-standard state
sub	ST_DISCONNECTED	{ 6 }	# not connected
# Operational modes
sub	MODE_EXTERNAL	{ 0 }	# using the external authentification
sub	MODE_INTERNAL	{ 1 }	# using the internal socket pre-authentified
# Requests for select()ing
sub	REQ_RD_START	{ 0 }	# socket must be select()ed for reading
sub	REQ_RD_STOP	{ 1 }	# socket must not be select()ed for reading
sub	REQ_WR_START	{ 2 }	# socket must be select()ed for writing
sub	REQ_WR_STOP	{ 3 }	# socket must not be select()ed for writing

#
# Internal constants
#

# Length blocks read/written on the socket
sub	IO_BLK_LEN	{ 2048 }
# Pathnames
sub	PATH_NS_USER_CMD	{ "$ENV{HOME}/.ns/ns_user_cmd" }
sub	PATH_NS_USER_CMD_TARGET	{ "/usr/local/libexec/ns_user_cmd" }
sub	PATH_NS_USER_LINK	{ "$ENV{HOME}/.ns/$ENV{NS_USER_LINK}" }

# ====================================================================
# Constructor. Always returns the object(success).
#
# Options required (unless the contrary is specified):
#   login:		login to be authentified
#   passwd:		passwd used to authentify the login
#   ns_host:		netsoul host
#   ns_port:		netsoul port
#   local_addr:		local host bind address
#   location:		location
#   data:		user data
#   state_st:		state (one of the &ST_* constants)
#   event_handler:	called when the module wants to notify about an event
#   request_handler:	called when the module wants the user start or stop
#			select()ing the socket for reading or writing (it is
#			assumed that at the begining, the user doesn't have to
#			select() the socket)
#   log_handler:	(optional) called when the module wants to give misc
#			information; if undef, msgs will be printed on stdout
#
# All the handlers above are a reference to an array: the 1st element is a
# reference to a sub, the other elements are its first args. The sub will be
# called with these first args and one ore more extra args, depending on the
# handler type.
#
# The event handler can be called with these extra args (<std_fields> denotes:
# id, agent_authlvl, user_authlvl, login, ip, ip_infos, hosttype, location,
# group):
#   &EV_CONN,     (conn_state(&CONN_*))
#   &EV_LOGIN,    (<std_fields>, entity)
#   &EV_LOGOUT,   (<std_fields>, entity)
#   &EV_MSG,      (<std_fields>, entity, msg)
#   &EV_STATE,    (<std_fields>, entity, state, state_st(&ST_*), state_time)
#   &EV_TALK,     (<std_fields>, entity, talkmsg)
#   &EV_TALKANS,  (<std_fields>, entity, talkans)
#   &EV_NEWMAIL,  (<std_fields>, entity, from)
#   &EV_PING,     (ping_time)
#
# The request handler can be called with these extra args:
#   &REQ_RD_START
#   &REQ_RD_STOP
#   &REQ_WR_START
#   &REQ_WR_STOP
#
# The log handler can be called with these extra args:
#   "misc string"
#
sub	new
{
    my $class = shift;
    my %opt = @_;
    my $self = { };
    
    bless($self, $class);

    # external settings
    $self->{login} = $opt{-login};
    $self->{passwd} = $opt{-passwd};
    $self->{ns_host} = $opt{-ns_host};
    $self->{ns_port} = $opt{-ns_port};
    $self->{local_addr} = $opt{-local_addr};
    $self->{location} = ns_encode($opt{-location});
    $self->{data} = ns_encode($opt{-data});
    $self->{state_st} = ($opt{-state_st} == &ST_ACTIVE
	|| $opt{-state_st} == &ST_IDLE || $opt{-state_st} == &ST_LOCK
	|| $opt{-state_st} == &ST_CONNECTION || $opt{-state_st} == &ST_SERVER)
    ? $opt{-state_st} : &ST_SERVER;
    $self->{event_handler_sub} = shift @{$opt{-event_handler}};
    $self->{event_handler_args} = $opt{-event_handler};
    $self->{request_handler_sub} = shift @{$opt{-request_handler}};
    $self->{request_handler_args} = $opt{-request_handler};
    $self->{log_handler_sub} = shift @{$opt{-log_handler}};
    $self->{log_handler_args} = $opt{-log_handler};

    # internal stuff
    $self->{ns_user_cmd_target} = undef;
    # what is found in the banner
    $self->{id} = undef;
    $self->{challenge} = undef;
    $self->{local_ip} = undef;
    $self->{local_port} = undef;
    $self->{conn_time} = undef;
    # delta time between server and client
    $self->{dtime_server_client} = undef;
    $self->{op_mode} = get_op_mode();
    $self->{last_errmsg} = undef;
    $self->{buffer_rd} = '';
    $self->{buffer_wr} = '';
    # ref to an array of refs to arrays containing (sub, args, ...)
    $self->{list_users_handler_fifo} = [];
    # ref to an array of refs to hashes containing (login => "", id => n, ...)
    $self->{list_users_results} = [];
    $self->{conn_state} = &CONN_DOWN;
    $self->{sock} = undef;
    $self->{nb_connect_attempts} = 0;
    return $self;
}

# ====================================================================
# Externally accessible subs:
#   get_socket_handle
#   get_conn_state
#   get_nb_connect_attempts
#   get_server_time
#   connect
#   disconnect
#   read_socket		| the user must call these 2 subs, depending
#   write_socket	| on the {read,writ}ability of the socket
#   list_users
#   send_msg
#   watch_log
#   send_ping
#   set_data
#   set_state
#

#
# Get the socket handle, the connection state, the number of connection
# attempts and the server time.
#
sub	get_socket_handle	{ return (shift)->{sock} }
sub	get_conn_state		{ return (shift)->{conn_state} }
sub	get_nb_connect_attempts	{ return (shift)->{nb_connect_attempts} }
sub	get_server_time		{ return time + (shift)->{dtime_server_client}}

#
# Connect the socket. Returns 0(failure) or 1(success).
#
sub	connect
{
    my $self = shift;

    # sanitary checks
    if ($self->{conn_state} != &CONN_DOWN) {
	$self->log("connection state is '" . str_of_conn($self->{conn_state})
	    . "', won't reconnect\n");
	return 0;
    }
    # the user can use nb_connect_attempts to delay the reconnection
    $self->{nb_connect_attempts}++;
    # reset the last error msg
    $self->{last_errmsg} = undef;
    # IO::Socket::INET is not used because the 'Blocking' option does
    # not seems to work, so the low-level OO interface IO::Socket is used
    $self->{sock} = IO::Socket->new;
    if ($self->{op_mode} == &MODE_EXTERNAL) {
	if (!defined ($self->{sock}->socket
		(PF_INET, SOCK_STREAM, scalar getprotobyname('tcp')))) {
	    $self->log("can't create PF_INET/SOCK_STREAM socket: $!\n");
	    return 0;
	}
    } else {
	if (!defined ($self->{sock}->socket(PF_UNIX, SOCK_STREAM, 0))) {
	    $self->log("can't create PF_UNIX/SOCK_STREAM socket: $!\n");
	    return 0;
	}
    }
    if (!defined ($self->{sock}->blocking(0))) {
	$self->log("can't set non-blocking mode on socket: $!\n");
	return 0;
    }
    # Note: when IO::Socket::connect() encounters the error EINPROGRESS, it can
    # returns either undef (IO::Socket 1.26) or the socket (IO::Socket 1.27),
    # that's why this particular error condition is tested
    if ($self->{op_mode} == &MODE_EXTERNAL) {
	my $iaddr;
	if (!defined ($iaddr = inet_aton($self->{ns_host}))) {
	    $self->log("can't resolve '$self->{ns_host}'\n");
	    return 0;
	}
	if (!defined ($self->{sock}->connect(scalar sockaddr_in
		    ($self->{ns_port}, $iaddr)))
	    && !$!{EINPROGRESS}) {
	    $self->log("can't connect to '$self->{ns_host}:$self->{ns_port}': "
		. "$!\n");
	    return 0;
	}
    } else {
	if (!defined ($self->{sock}->connect(scalar sockaddr_un
		    (&PATH_NS_USER_LINK)))
	    && !$!{EINPROGRESS}) {
	    $self->log("can't connect to '" . &PATH_NS_USER_LINK . "': $!\n");
	    return 0;
	}
    }
    # We set conn_state to CONN_TRY only now (and not at the beginning of
    # connect()) because if connect() would failed, conn_state would have
    # to be reset to CONN_DOWN, the event EV_CONN/CONN_DOWN would be
    # immediately generated, the user would probably try to re-connect()
    # in the event handler (called by connect), leading to the same failure:
    # an infinite loop would be initiated !
    $self->set_conn_state(&CONN_TRY);
    # The socket is in non-blocking mode, it must be select()ed for writing,
    # then when write_socket() is called, the success or failure of the
    # connection will be evaluated. See connect(2) for more details.
    $self->{request_handler_sub}(@{$self->{request_handler_args}},
	&REQ_WR_START);
    return 1;
}

#
# Disconnect the socket. Always returns 1(success).
#
# Warning: whenever this sub is called (by the Netsoul module or by the
# user), an EV_CONN/CONN_DOWN event will be generated. Take this into
# account, especially when connect() is called on this event to handle
# disconnection errors.
#
sub	disconnect
{
    my $self = shift;

    return 1 if ($self->{conn_state} == &CONN_DOWN);

    # reenable other clients when the connection shuts down
    $self->reenable_other_clients;

    # call the request handler to simplify the user's life (he won't
    # have to reset the requests himself on EV_CONN/CONN_DOWN events)
    $self->{request_handler_sub}(@{$self->{request_handler_args}},
	&REQ_RD_STOP);
    $self->{request_handler_sub}(@{$self->{request_handler_args}},
	&REQ_WR_STOP);

    # reinit internal datas
    $self->{buffer_rd} = '';
    $self->{buffer_wr} = '';
    $self->{list_users_handler_fifo} = [];
    $self->{list_users_results} = [];

    # in internal mode, we try to delete the watch_log list, we don't handle
    # the cases where syswrite fails to write the whole datas because:
    # (1) it is not worth bothering to use a complex system such as the
    #     requests for {read,writ}ing
    # (2) we want disconnect() to be *fast*, so we call syswrite() and if it
    #     fails, never mind !
    # the leading \012 is here to start a new cmd (the previous cmd might not
    # have been completely written since the user can call disconnect() at
    # any time)
    $self->{sock}->syswrite("\012cmd watch_log_user {}\012")
    if ($self->{op_mode} == &MODE_INTERNAL);

    # shutdown the connection and automatically closes the socket if it's open
    $self->{sock}->shutdown(&SHUT_RDWR);
    undef $self->{sock};

    # this should be the only point where set_conn_state(&CONN_DOWN) is called
    $self->set_conn_state(&CONN_DOWN);
    return 1;
}

#
# Read the socket. It must be called by the user of the Netsoul module
# when the socket is readable but *only* when a request for select()ing
# it for reading has been issued by the Netsoul module. Always returns
# 1(success) because it may be required by the user code (eg: when this
# sub is called with the Tk::fileevent mecanism).
#
sub	read_socket
{
    my $self = shift;

    if ($self->{conn_state} == &CONN_UP
	|| $self->{conn_state} == &CONN_AUTH_REQ
	|| $self->{conn_state} == &CONN_LOG_REQ
	|| $self->{conn_state} == &CONN_READY) {
	# read the banner || read the auth_ag reply ||
	# read the ext_user_log reply || read any events
	my ($buf, $nread);
	if (!defined ($nread = $self->{sock}->sysread($buf, &IO_BLK_LEN))
		or !$nread) {
	    if (!defined ($nread)) { $self->log("socket read error: $!\n") }
	    else { $self->log("connection closed by server\n") }
	    return $self->disconnect;
	}
	$self->log("recv [" . (chomp($_=$buf), $_) . "]\n");
	$self->{buffer_rd} .= $buf;
	if ($self->{buffer_rd} =~ /\012/) {
	    if ($self->{conn_state} == &CONN_UP) {
		$self->{request_handler_sub}(@{$self->{request_handler_args}},
		    &REQ_RD_STOP);
		$self->set_conn_state(&CONN_BANNER);
		return $self->disconnect if (!$self->extract_and_check_banner);
		$self->{buffer_wr} .= "auth_ag ext_user none none\012";
		$self->{request_handler_sub}(@{$self->{request_handler_args}},
		    &REQ_WR_START);
	    } elsif ($self->{conn_state} == &CONN_AUTH_REQ) {
		$self->{request_handler_sub}(@{$self->{request_handler_args}},
		    &REQ_RD_STOP);
		$self->set_conn_state(&CONN_AUTH_REP);
		return $self->disconnect if (!$self->extract_and_check_reply);
		$self->{buffer_wr} .= "ext_user_log $self->{login} "
		. md5_hex($self->{challenge} . "-" . $self->{local_ip} . "/" .
		    $self->{local_port} . $self->{passwd})
		. " $self->{location} $self->{data}\012";
		$self->{request_handler_sub}(@{$self->{request_handler_args}},
		    &REQ_WR_START);
	    } elsif ($self->{conn_state} == &CONN_LOG_REQ) {
		$self->set_conn_state(&CONN_LOG_REP);
		return $self->disconnect if (!$self->extract_and_check_reply);
		# set connection state to CONN_READY
		$self->set_conn_state_to_ready;
	    } elsif ($self->{conn_state} == &CONN_READY) {
		# in others CONN_* one need to read only 1 line at a time, but
		# in CONN_READY one can receive more than 1 line at a time
		while ($self->{buffer_rd} =~ /\012/) {
		    $self->extract_and_check_event;
		}
	    }
	}
    } else {
	$self->log("read_socket: should not be called while in state '"
	    . str_of_conn($self->{conn_state}) . "'\n");
    }
    return 1;
}

#
# read_socket() comment's =~ s/read/write/;
#
sub	write_socket
{
    my $self = shift;

    if ($self->{conn_state} == &CONN_TRY) {
	# the connect() was non-blocking, one has the result of connect()
	$self->{request_handler_sub}(@{$self->{request_handler_args}},
	    &REQ_WR_STOP);
	if ($_ = $self->{sock}->sockopt(SO_ERROR)) {
	    $! = $_;
	    $self->log("connecting to '"
		. ($self->{op_mode} == &MODE_EXTERNAL ?
		    "$self->{ns_host}:$self->{ns_port}" : &PATH_NS_USER_LINK)
		. "' has failed: $!\n");
	    return $self->disconnect;
	} else {
	    # in external mode one have to read the banner etc, while
	    # in internal mode the connection is directly usable
	    if ($self->{op_mode} == &MODE_EXTERNAL) {
		$self->set_conn_state(&CONN_UP);
	    } else {
		# this sets the delta time (only for the internal mode);
		# one supposes it is 0 because internal hosts normally have
		# their clock synchronised with the server one)
		$self->{dtime_server_client} = 0;
		# set connection state to CONN_READY
		$self->set_conn_state_to_ready;
	    }
	    # one wants to read the banner or to listen to events
	    $self->{request_handler_sub}(@{$self->{request_handler_args}},
		&REQ_RD_START);
	}
    } elsif ($self->{conn_state} == &CONN_BANNER
	|| $self->{conn_state} == &CONN_AUTH_REP
	|| $self->{conn_state} == &CONN_READY) {
	# write auth_ag || write ext_user_log || write misc cmd
	my $nwritten;
	if (!defined ($nwritten = $self->{sock}->syswrite
		($self->{buffer_wr}, &IO_BLK_LEN))) {
	    $self->log("socket write error: $!\n");
	    return $self->disconnect;
	}
	$self->log("sent ["
	    . (chomp($_=substr($self->{buffer_wr}, 0, $nwritten)), $_) . "]\n");
	$self->{buffer_wr} = substr($self->{buffer_wr}, $nwritten);
	if ($self->{buffer_wr} eq '') {
	    $self->{request_handler_sub}(@{$self->{request_handler_args}},
		&REQ_WR_STOP);
	    if ($self->{conn_state} == &CONN_BANNER) {
		$self->set_conn_state(&CONN_AUTH_REQ);
		# one wants to read the auth_ag reply
		$self->{request_handler_sub}(@{$self->{request_handler_args}},
		    &REQ_RD_START);
	    } elsif ($self->{conn_state} == &CONN_AUTH_REP) {
		$self->set_conn_state(&CONN_LOG_REQ);
		# one wants to read the ext_user_log reply
		$self->{request_handler_sub}(@{$self->{request_handler_args}},
		    &REQ_RD_START);
	    } # nothing to do in the case CONN_READY
	}
    } else {
	$self->log("write_socket: should not be called while in state '"
	    . str_of_conn($self->{conn_state}) . "'\n");
    }
    return 1;
}

#
# Interface to list_users and cmd who. Returns 1(success) or 0(failure).
# Arguments:
#   -res_handler => [sub,args]	(mandatory, where to return results)
#   -logins => ["log0","log1"]	(optional, defaults to <undef>: list all users)
#
# The res_handler is a reference to an array, the 1st element is a reference
# to a sub, the other elements are its first args; the sub will be called
# (with its first args and one extra arg) when all the reply lines have been
# received, the extra arg will be an array of references to hashes containings
# fields similar to the one used by the event handler of the Netsoul object:
#
#   <std_fields>, login_time, idle_time, data, state, state_st, state_time
#
# Example of a valid res_handler:
#   -res_handler => [\&res_handler, "res_handler() called"]
# with:
#   sub res_handler { print shift; print "$_->{login}\n" foreach (@_) }
#
# The list_users cmd is used whenever possible (that is in external mode)
# because its replies consume less bandwith (no "user_cmd ... | who " prefix)
# and it is more secure that cmd who (it is possible to generate a fake
# cmd who reply and not being able to differenciate it from a real reply).
#
sub	list_users
{
    my $self = shift;
    my %opt = @_;

    if ($self->{conn_state} != &CONN_READY) {
	$self->log("list_users: connection not ready\n"); return 0;
    }
    if (!defined $opt{-res_handler}) {
	$self->log("list_users: result handler not specified\n"); return 0;
    }
    push @{$self->{list_users_handler_fifo}}, $opt{-res_handler};
    $self->{buffer_wr} .=
    ($self->{op_mode} == &MODE_EXTERNAL ? "list_users " : "cmd who ")
    . (defined $opt{-logins} ? "{" . join(',', @{$opt{-logins}}) . "}" : "")
    . "\012";
    $self->{request_handler_sub}(@{$self->{request_handler_args}},
	&REQ_WR_START);
    return 1;
}

#
# Interface to cmd msg_user. Returns 1(success) or 0(failure). Arguments:
#   -login => "login"	(or -id)
#   -id => 42		(or -login)
#   -msg => "hello"	(optional, defaults to "my name is geronimo")
#
sub	send_msg
{
    my $self = shift;
    my %opt = @_;
    my $dest = undef;

    if ($self->{conn_state} != &CONN_READY) {
	$self->log("send_msg: connection not ready\n"); return 0;
    }
    $dest = $opt{-login} if (defined $opt{-login});
    $dest = ":" . $opt{-id} if (defined $opt{-id});
    if (!defined $dest) {
	$self->log("send_msg: nor -login neither -id specified\n"); return 0;
    }
    $self->{buffer_wr} .= ($self->{op_mode} == &MODE_EXTERNAL ? "user_" : "")
    . "cmd msg_user $dest msg "
    .  (defined $opt{-msg} ? ns_encode($opt{-msg}) : "my name is geronimo")
    . "\012";
    $self->{request_handler_sub}(@{$self->{request_handler_args}},
	&REQ_WR_START);
    return 1;
}

#
# Interface to cmd watch_log_user. Returns 1(success) or 0(failure). Arguments:
#   -logins => ["log0","log1"]	(optional, defaults to []: watch_log no user)
#
sub	watch_log
{
    my $self = shift;
    my %opt = @_;

    if ($self->{conn_state} != &CONN_READY) {
	$self->log("watch_log: connection not ready\n"); return 0;
    }
    $self->{buffer_wr} .= ($self->{op_mode} == &MODE_EXTERNAL ? "user_" : "")
    . "cmd watch_log_user "
    . (defined $opt{-logins} ? "{" . join(',', @{$opt{-logins}}) . "}" : "{}")
    . "\012";
    $self->{request_handler_sub}(@{$self->{request_handler_args}},
	&REQ_WR_START);
    return 1;
}

#
# Interface to ping. Returns 1(success) or 0(failure).
#
sub	send_ping
{
    my $self = shift;

    if ($self->{conn_state} != &CONN_READY) {
	$self->log("send_ping: connection not ready\n"); return 0;
    }
    $self->{buffer_wr} .= "ping\012";
    $self->{request_handler_sub}(@{$self->{request_handler_args}},
	&REQ_WR_START);
    return 1;
}

#
# Interface to cmd user_data. Returns 1(success) or 0(failure). Arguments:
#   -data => "geronimo-cvs"	(optional, defaults to "-")
#
sub	set_data
{
    my $self = shift;
    my %opt = @_;

    if ($self->{conn_state} != &CONN_READY) {
	$self->log("set_data: connection not ready\n"); return 0;
    }
    $self->{buffer_wr} .= ($self->{op_mode} == &MODE_EXTERNAL ? "user_" : "")
    . "cmd user_data " . (defined $opt{-data} ? ns_encode($opt{-data}) : "-")
    . "\012";
    $self->{request_handler_sub}(@{$self->{request_handler_args}},
	&REQ_WR_START);
    return 1;
}

#
# Interface to cmd state. Returns 1(success) or 0(failure). Arguments:
#   -state_st => &ST_*	(optional, defaults to &ST_SERVER)
#
sub	set_state
{
    my $self = shift;
    my %opt = @_;
    my $state;

    if ($self->{conn_state} != &CONN_READY) {
	$self->log("set_state: connection not ready\n"); return 0;
    }
    $state = "server";
    ($opt{-state_st} == &ST_ACTIVE)	and $state = "actif";
    ($opt{-state_st} == &ST_IDLE)	and $state = "idle";
    ($opt{-state_st} == &ST_LOCK)	and $state = "lock";
    ($opt{-state_st} == &ST_CONNECTION)	and $state = "connection";
    ($opt{-state_st} == &ST_SERVER)	and $state = "server";
    $self->{buffer_wr} .= ($self->{op_mode} == &MODE_EXTERNAL ? "user_" : "")
    . "cmd state " . $state . ":" . $self->get_server_time . "\012";
    $self->{request_handler_sub}(@{$self->{request_handler_args}},
	&REQ_WR_START);
    return 1;
}

# ====================================================================
# Internal subs:
#   disable_other_clients
#   reenable_other_clients
#   extract_and_check_banner
#   extract_and_check_reply
#   extract_and_check_event
#   set_conn_state
#   set_conn_state_to_ready
#   log
#

#
# Disable other netsoul clients. This is called when the connection becomes
# ready.
#
sub	disable_other_clients
{
    my $self = shift;

    # this sub don't have to be executed in external mode
    return if ($self->{op_mode} == &MODE_EXTERNAL);

    # remove the ns_user_cmd link
    $_ = &PATH_NS_USER_CMD;
    if (-e) {
	if (-l) {
	    my $target = readlink;
	    my $errmsg = "$!";
	    if (1 == unlink) {
		if (defined $target) {
		    $self->log("the symlink '$_' has been successfuly removed "
			. "(it was pointing to '$target'), it will be "
			. "recreated at connection shutdown\n");
		    $self->{ns_user_cmd_target} = $target;
		} else {
		    $self->log("the symlink '$_' has been successfuly removed "
			. "(but readlink failed ($errmsg), it will be supposed "
			. "that it was pointing to '" . &PATH_NS_USER_CMD_TARGET
			. "'), it will be recreated at connection shutdown\n");
		    $self->{ns_user_cmd_target} = &PATH_NS_USER_CMD_TARGET;
		}
	    } else {
		$self->log("the symlink '$_' can not be removed ($!), another "
		    . "netsoul client will probably be running at the same "
		    . "time that Geronimo\n");
	    }
	} else {
	    $self->log("the file '$_' exists but is not a symlink, Geronimo "
		. "don't know how to disable this netsoul client, it will "
		. "probably be running at the same time that Geronimo\n");
	}
    } # else there is probably no other netsoul client running

    # kill the ns_user_cmd process
    if (open(PS, "ps x |")) {
	my ($pid, $cmd);
	while (<PS>) {
	    if (/^(\d+)\s+\S+\s+\S+\s+\S+\s+(perl: ns_user_cmd )$/) {
		($pid, $cmd) = ($1, $2);
		1 == kill("TERM", $pid)
		? $self->log("'$cmd' (pid $pid) successfuly terminated\n")
		: $self->log("can't send SIGTERM to '$cmd' (pid $pid): $!\n");
		last;
	    }
	}
	close PS;
	$self->log("'ns_user_cmd' process not found\n") if (!defined $cmd);
    } else {
	$self->log("can't read from 'ps x |': $!\n");
    }
}

#
# Re-enable other netsoul clients. This is called when the connection
# shuts down.
#
sub	reenable_other_clients
{
    my $self = shift;

    # this sub don't have to be executed in external mode
    return if ($self->{op_mode} == &MODE_EXTERNAL);

    # recreate the symlink
    if (defined $self->{ns_user_cmd_target}) {
	if (symlink($self->{ns_user_cmd_target}, &PATH_NS_USER_CMD)) {
	    $self->log("the symlink '" . &PATH_NS_USER_CMD . "' pointing to '"
		. $self->{ns_user_cmd_target} . "' has been successfuly "
		. "recreated\n");
	} else {
	    $self->log("the symlink '" . &PATH_NS_USER_CMD . "' that should "
		. "points to '" . $self->{ns_user_cmd_target} . "' can't be "
		. "recreated ($!), another netsoul client will probably not "
		. "not run anymore\n");
	}
    }
}

#
# Parse the banner in the read buffer. Returns 1(success) or 0(failure).
#
sub	extract_and_check_banner
{
    my $self = shift;
    my $banner;

    if ($self->{buffer_rd} !~ /^(.*?)\012(.*)$/s) {
	$self->log("malformed banner in buffer '$self->{buffer_rd}'\n");
	return 0;
    }
    ($banner, $self->{buffer_rd}) = ($1, $2);

    if ($banner !~ /^salut\s+(\d+)\s+(\S+)\s+(\S+)\s+(\d+)\s+(\d+)\s*$/) {
	$self->log("invalid banner '$banner'\n");
	return 0;
    }
    $self->{id} = $1;
    $self->{challenge} = $2;
    $self->{local_ip} = $3;
    $self->{local_port} = $4;
    $self->{conn_time} = $5;
    # this sets the delta time (only for the external mode)
    $self->{dtime_server_client} = $self->{conn_time} - time;
    if (length $self->{challenge} < 32) {
	$self->log("challenge '$self->{challenge}' is anormaly short\n");
	return 0;
    }
    if ($self->{local_ip} !~ /^\d+(\.\d+){3}$/) {
	$self->log("'$self->{local_ip}' is not a valid ip\n");
	return 0;
    }
    return 1;
}

#
# Check if a cmd reply is "good". Returns 1(success) or 0(failure).
#
sub	extract_and_check_reply
{
    my $self = shift;
    my $rep;

    if ($self->{buffer_rd} !~ /^(.*?)\012(.*)$/s) {
	$self->log("malformed cmd reply in buffer '$self->{buffer_rd}'\n");
	return 0;
    }
    ($rep, $self->{buffer_rd}) = ($1, $2);

    if ($rep !~ /^rep 002 --/) {
	$self->log("reply '$rep' indicates an error\n");
	$self->{last_errmsg} = $rep;
	$self->{last_errmsg} .=
	" (invalid login/password)" if ($rep =~ /^rep 033 --/);
	return 0;
    }
    return 1;
}

#
# Extract and check an event frome the read buffer. Returns 1(success)
# or 0(failure).
#
sub	extract_and_check_event
{
    my $self = shift;
    my ($event, @res);

    if ($self->{buffer_rd} !~ /^(.*?)\012(.*)$/s) {
	$self->log("malformed event in buffer '$self->{buffer_rd}'\n");
	return 0;
    }
    ($event, $self->{buffer_rd}) = ($1, $2);

    $event =~ s/ $//; # a bogus extra space (added by ns-server) is removed
    if (0 < scalar (@res = parse_std_event($event))) {
	# standard event
	my (%hash, $cmd, $arg);
	($hash{id}, $hash{entity}, $hash{agent_authlvl}, $hash{user_authlvl},
	    $hash{login}, $hash{ip}, $hash{ip_infos}, $hash{hosttype},
	    $hash{location}, $hash{group}, $cmd, $arg) = @res;
	if ($cmd eq "login") {
	    $self->{event_handler_sub}(@{$self->{event_handler_args}},
		&EV_LOGIN, %hash);
	} elsif ($cmd eq "logout") {
	    $self->{event_handler_sub}(@{$self->{event_handler_args}},
		&EV_LOGOUT, %hash);
	} elsif ($cmd eq "msg") {
	    $hash{msg} = ns_decode($arg);
	    $self->{event_handler_sub}(@{$self->{event_handler_args}},
		&EV_MSG, %hash);
	} elsif ($cmd eq "state") {
	    ($hash{state}, $hash{state_st}, $hash{state_time}) =
	    parse_user_state($arg);
	    $self->{event_handler_sub}(@{$self->{event_handler_args}},
		&EV_STATE, %hash);
	} elsif ($cmd eq "talk") {
	    $hash{talkmsg} = ns_decode($arg);
	    $self->{event_handler_sub}(@{$self->{event_handler_args}},
		&EV_TALK, %hash);
	} elsif ($cmd eq "talkans") {
	    $hash{talkans} = $arg;
	    $self->{event_handler_sub}(@{$self->{event_handler_args}},
		&EV_TALKANS, %hash);
	} elsif ($cmd eq "new_mail") {
	    # ns_user_cmd check these conditions on new_mail events
	    if (($hash{agent_authlvl} > 5 || $hash{user_authlvl} > 5)
		&& ($hash{login} eq "_deamon")) {
		$arg =~ s/^-f //;
		$hash{from} = $arg;
		$self->{event_handler_sub}(@{$self->{event_handler_args}},
		    &EV_NEWMAIL, %hash);
	    } else {
		$self->log("untrusted '$cmd' cmd in event '$event'\n");
		return 0;
	    }
	} elsif ($cmd eq "who") {
	    # one line (or end) of a cmd who event
	    if ($hash{login} eq $self->{login}) {
		if (0 < scalar (@res = parse_list_users_line($arg))) {
		    %hash = ();
		    ($hash{id}, $hash{login}, $hash{ip}, $hash{ip_infos},
			$hash{login_time}, $hash{idle_time},
			$hash{user_authlvl}, $hash{agent_authlvl},
			$hash{hosttype}, $hash{location}, $hash{group},
			$hash{state}, $hash{state_st}, $hash{state_time},
			$hash{data}) = @res;
		    push @{$self->{list_users_results}}, \%hash;
		} elsif ($arg =~ /^rep 002 --/) {
		    my $handler = shift @{$self->{list_users_handler_fifo}};
		    my $handler_sub = shift @{$handler};
		    &$handler_sub(@{$handler}, @{$self->{list_users_results}});
		    $self->{list_users_results} = [];
		} else {
		    $self->log("malformed cmd who event '$event'\n");
		    return 0;
		}
	    } else {
		$self->log("untrusted source in cmd who event '$event'\n");
		return 0;
	    }
	} else {
	    $self->log("unknown '$cmd' cmd in standard event '$event'\n");
	    return 0;
	}
    } elsif (0 < scalar (@res = parse_list_users_line($event))) {
	# one line of a list_users event
	my (%hash);
	($hash{id}, $hash{login}, $hash{ip}, $hash{ip_infos}, $hash{login_time},
	    $hash{idle_time}, $hash{user_authlvl}, $hash{agent_authlvl},
	    $hash{hosttype}, $hash{location}, $hash{group}, $hash{state},
	    $hash{state_st}, $hash{state_time}, $hash{data}) = @res;
	push @{$self->{list_users_results}}, \%hash;
    } elsif ($event =~ /^rep 002 --/) {
	# end of a list_users event
	my $handler = shift @{$self->{list_users_handler_fifo}};
	my $handler_sub = shift @{$handler};
	&$handler_sub(@{$handler}, @{$self->{list_users_results}});
	$self->{list_users_results} = [];
    } elsif ($event =~ /^ping (\d+)$/) {
	# ping event
	$self->{event_handler_sub}(@{$self->{event_handler_args}},
	    &EV_PING, (ping_time => $1));
    } else {
	$self->log("unrecognized event '$event'\n");
	return 0;
    }
    return 1;
}

#
# Set the connection state and generate the associated event.
#
sub	set_conn_state
{
    my $self = shift;
    my ($st) = @_;

    $self->{conn_state} = $st;
    $self->{event_handler_sub}(@{$self->{event_handler_args}},
	&EV_CONN, (conn_state => $st));
}

#
# Set the connection state to CONN_READY, and do the stuff that has to
# be done when the connection becomes ready.
#
sub	set_conn_state_to_ready
{
    my $self = shift;

    # conn_state is manually set to CONN_READY to make set_state() happy
    # then a CONN_READY event is generated (this should be the only point
    # where set_conn_state(&CONN_READY) is called)
    $self->{conn_state} = &CONN_READY;
    $self->set_state(-state_st => $self->{state_st});
    $self->set_conn_state(&CONN_READY);
    # disable other clients when the connection becomes ready
    $self->disable_other_clients;
}

#
# Call the log handler if necessary.
#
sub	log
{
    my $self = shift;
    my ($msg) = @_;

    if (defined $self->{log_handler_sub}) {
	$self->{log_handler_sub}(@{$self->{log_handler_args}}, $msg);
    } else {
	print $msg;
    }
}

# ====================================================================
# Misc subs that don't need the object reference ($self)
#   get_op_mode
#   str_of_conn
#   parse_std_event
#   parse_list_users_line
#   parse_user_state
#   ip_extra_infos
#   ns_decode
#   ns_encode
#

#
# Get the operational mode.
#
sub	get_op_mode
{
    defined $ENV{NS_USER_LINK} ? &MODE_INTERNAL : &MODE_EXTERNAL;
}

#
# Get a string that describe the connection state. It should be only used
# to display a friendly &CONN_* representation to a user.
#
sub	str_of_conn
{
    my ($conn_state) = @_;

    ($conn_state == &CONN_DOWN)		and return "not connected";
    ($conn_state == &CONN_TRY)		and return "trying to connect";
    ($conn_state == &CONN_UP)		and return "connection established";
    ($conn_state == &CONN_BANNER)	and return "banner received";
    ($conn_state == &CONN_AUTH_REQ)	and return "auth_ag sent";
    ($conn_state == &CONN_AUTH_REP)	and return "auth_ag reply received";
    ($conn_state == &CONN_LOG_REQ)	and return "log cmd sent";
    ($conn_state == &CONN_LOG_REP)	and return "log cmd reply received";
    ($conn_state == &CONN_READY)	and return "connection ready";
    return "<unknown connection state $conn_state>";
}

#
# Parse a standard event.
#
sub	parse_std_event
{
    my ($event) = @_;
    my ($id, $entity, $agent_authlvl, $user_authlvl, $login_hostname,
	$ip, $ip_infos, $hosttype, $location, $group, $cmd, $arg);


    if ($event =~ /^(?:user_|)cmd\s+([^:]*):([^:]*):([^:]*):([^:]*):([^:]*):(.*?)\s+\|\s+(.*)$/) {
	# standard event
	$id = $1;
	$entity = $2;
	($agent_authlvl, $user_authlvl) = (split /\//, $3);
	($login_hostname, $ip) = (split /\@/, $4); $ip ||= "<n/a>";
	$ip_infos = ($ip ne "<n/a>") ? ip_extra_infos($ip) : "<n/a>";
	$hosttype = $5;
	$location = $6;
	$cmd = $7;

	# finish parsing of some "complex" fields.

	# below, we rely on the "greediness" of regular expressions to parse
	# the location because it can contain a colon ':'
	if ($location =~ /(.*):(.*)/) {
	    $location = ns_decode($1);
	    $group = $2;
	} else {
	    $location = ns_decode($location);
	    $group = "<n/a>";
	}
	$cmd =~ /^(\S+)\s*(.*)$/; # \s* because not all cmds have an arg
	($cmd, $arg) = ($1, $2);
	return ($id, $entity, $agent_authlvl, $user_authlvl, $login_hostname,
	    $ip, $ip_infos, $hosttype, $location, $group, $cmd, $arg);
    } else {
	return ();
    }
}

#
# Parse a line as returned by the list_users command.
#
sub	parse_list_users_line
{
    my ($event) = @_;
    my ($id, $login, $ip, $ip_infos, $login_time, $idle_time, $user_authlvl,
	$agent_authlvl, $hosttype, $location, $group, $state, $state_st,
	$state_time, $data);

    if ($event =~ /^(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)$/) {
	($id, $login, $ip, $login_time, $idle_time, $user_authlvl,
	    $agent_authlvl, $hosttype, $location, $group, $state, $data) =
	($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12);
	$ip_infos = ip_extra_infos($ip);
	($state, $state_st, $state_time) = parse_user_state($state);
	$location = ns_decode($location);
	$data = ns_decode($data);
	return ($id, $login, $ip, $ip_infos, $login_time, $idle_time,
	    $user_authlvl, $agent_authlvl, $hosttype, $location, $group,
	    $state, $state_st, $state_time, $data);
    } else {
	return ();
    }
}

#
# Parse a user state string (eg: "idle:1234", "connection") and returns a
# (str, &ST_*, num) 3-uplet.
#
sub	parse_user_state
{
    my ($state_str) = @_;
    my ($state, $state_st, $state_time);

    if ($state_str =~ /(.*):(\d+)$/) {
	($state, $state_time) = ($1, $2);
    } else {
	($state, $state_time) = ($state_str, undef);
    }
    for ($state) {
	/^actif$/	and $state_st = &ST_ACTIVE, last;
	/^(idle|away)$/	and $state_st = &ST_IDLE, last;
	/^lock$/	and $state_st = &ST_LOCK, last;
	/^connection$/	and $state_st = &ST_CONNECTION, last;
	/^server$/	and $state_st = &ST_SERVER, last;
	$state_st = &ST_UNKNOWN;
    }
    return ($state, $state_st, $state_time);
}

#
# Return extra information about an IP.
#
sub	ip_extra_infos
{
    my ($ip) = @_;
    my %ip_table =
    (
	"10.251." =>		"lab-cisco-mid-sr",
	"10.227.42." =>		"lse",
	"10.227.4." =>		"sda",
	"10.227." =>		"lab",
	"10.226.7." =>		"lab-tcom",
	"10.226.6." =>		"lab-acu",
	"10.226.5." =>		"lab-console",
	"10.226." =>		"lab-mspe",
	"10.225.19." =>		"epitanim",
	"10.225.18." =>		"epidemic",
	"10.225.10." =>		"sda",
	"10.225.8." =>		"cycom",
	"10.225.7." =>		"epx",
	"10.225.6." =>		"prologin",
	"10.225.2." =>		"nomad",
	"10.225." =>		"assos",
	"10.224.14." =>		"sda",
	"10.223.106." =>	"www",
	"10.223.104." =>	"episport",
	"10.223.103." =>	"epicom",
	"10.223.100." =>	"bde-epita",
	"10.223.42." =>		"omatis",
	"10.223.15." =>		"ipsa",
	"10.223.13." =>		"lrde",
	"10.223.7." =>		"cvi",
	"10.223.1." =>		"epi",
	"10.223." =>		"pasteur",
	"10.42.42." =>		"bocal",
	"10.42." =>		"sm",
	"10.1." =>		"adm",
	"10." =>		"epita",
    );

    ($ip =~ /^\Q$_\E/ and return $ip_table{$_})
    foreach (sort {$b cmp $a} keys %ip_table);
    # %ip_table keys are sorted in reversed order so that IPs will be
    # first matched with "x.y.z." then with "x.y." then with "x."
    return "ext";
}

#
# "Netsoul-decode" a string.
#
sub	ns_decode
{
    my ($str) = @_;
    
    $str =~ tr/+/ /;
    $str =~ s/%([0-9a-fA-F]{2})/pack("c", hex($1))/ge;
    $str =~ s[\\n][\n]g;
    return $str;
}

#
# "Netsoul-encode" a string.
#
sub	ns_encode
{
    my ($str) = @_;
    
    $str =~ s[\n][\\n]g;
    $str =~ s/([^a-zA-Z0-9_.\-\\])/sprintf("%%%02X", ord($1))/ge;
    $str =~ tr/ /+/;
    return $str;
}

1;
